@@ -113,6 +113,8 @@ void parse_configfile(const char *cmdpath, entropy_source_t **entropy_source_p)
       entropy_source[cur_cmd].args[arg] = s;
       arg++;
     } while ((arg < 5) && (cp = strtok(NULL, WHITESPACE)));
+    while (arg < 5)
+      entropy_source[cur_cmd].args[arg++] = NULL;
 
     if (strtok(NULL, WHITESPACE))
        fprintf(stderr, "ignored extra command elements (max 5), %.100s:%d",
