@@ -8,6 +8,7 @@
 #include <time.h>
 #include <assert.h>
 #include <syslog.h>
+#include <pwd.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/un.h>
@@ -165,7 +166,7 @@ static void killpeer(int service_socket)
   fprintf(stderr, "Killing: %d\n", peer_id);
 #endif
 
-  if (kill(peer_id, SIGHUP) < 0)
+  if (kill(peer_id, SIGTERM) < 0)
   {
 #ifdef PID_T_IS_LONG_INT
     fprintf(stderr, "Cannot kill %ld: %s\n", peer_id, strerror(errno));
@@ -181,8 +182,8 @@ static void usage(const char *progname)
 {
   fprintf(stderr, "prngd %s (%s)\n", PRNGD_VERSION, PRNGD_DATE);
   fprintf(stderr,
-	  "Usage: %s [options] /path/to/socket1 [/path/to/socket2 ...]\nOptions:\n\t-d/--debug: debugging on\n\t-c/--cmdfile cmdpath: use cmdpath for entropy commands [%s]\n\t-s/--seedfile seedpath: use seedpath as seedfile [%s]\n\t-n/--no-seedfile: no seedfile, keep pool in memory only\n\t-k/--kill: kill daemon on other side\n\t-v/--version: print version and exit\n",
-	  progname, CONFIGFILE, RANDSAVENAME);
+	  "Usage: %s [options] /path/to/socket1 [/path/to/socket2 ...]\nOptions:\n\t-d/--debug: debugging on\n\t-c/--cmdfile cmdpath: use cmdpath for entropy commands [%s]\n\t-s/--seedfile seedpath: use seedpath as seedfile [%s]\n\t-n/--no-seedfile: no seedfile, keep pool in memory only\n\t-m/--mode mode: use mode for sockets [%04o]\n\t-k/--kill: kill daemon on other side\n\t-v/--version: print version and exit\n",
+	  progname, CONFIGFILE, RANDSAVENAME, SOCKETMODE);
 }
 
 int main(int argc, char *argv[])
@@ -198,9 +199,12 @@ int main(int argc, char *argv[])
   char **randsock = NULL;
   char *cmdpath = CONFIGFILE;
   char *seedpath = RANDSAVENAME;
+  unsigned int socketmode = SOCKETMODE;
   int *service_socket, kill_socket;
   struct sockaddr_un sockun, sockun_kill;
   struct stat randsock_stat;
+  struct passwd *pwent;
+  char *username = "unknown";
   entropy_source_t *entropy_source;
 
   /*
@@ -221,6 +225,15 @@ int main(int argc, char *argv[])
     else if ((!strcmp(argv[argp], "-s") || !strcmp(argv[argp], "--seedfile"))
 	     && (argp + 1 < argc))
       seedpath = argv[++argp];
+    else if ((!strcmp(argv[argp], "-m") || !strcmp(argv[argp], "--mode"))
+	     && (argp + 1 < argc))
+    {
+      if (sscanf(argv[++argp], "%o", &socketmode) == 0)
+      {
+	fprintf(stderr, "Cannot interprete socketmode '%s'\n", argv[argp]);
+	exit(EXIT_FAILURE);
+      }
+    }
     else if (!strcmp(argv[argp], "-n") || !strcmp(argv[argp], "--no-seedfile"))
       seedpath = NULL;
     else if (!strcmp(argv[argp], "-k") || !strcmp(argv[argp], "--kill"))
@@ -401,6 +414,12 @@ int main(int argc, char *argv[])
 	      strerror(errno));
       exit(EXIT_FAILURE);
     }
+    if (chmod(randsock[i], socketmode) < 0)
+    {
+      fprintf(stderr, "Could not chmod socket to %04o: %s\n", socketmode,
+	      strerror(errno));
+      exit(EXIT_FAILURE);
+    }
     if (listen(service_socket[i], SOMAXCONN) < 0)
     {
       fprintf(stderr, "Could not listen on socket: %s\n", strerror(errno));
@@ -411,18 +430,7 @@ int main(int argc, char *argv[])
      * hang inside accept() if a connection is closed between select() and
      * accept().
      */
-    if ((flags = fcntl(service_socket[i], F_GETFL, 0)) < 0)
-    {
-      fprintf(stderr, "Could not get socket file mode: %s\n",
-	      strerror(errno));
-      exit(EXIT_FAILURE);
-    }
-    if (fcntl(service_socket[i], F_SETFL, flags | O_NONBLOCK) < 0)
-    {
-      fprintf(stderr, "Could not set socket file mode: %s\n",
-	      strerror(errno));
-      exit(EXIT_FAILURE);
-    }
+    non_blocking(service_socket[i], 1);
     close_on_exec(service_socket[i], 1);
   }
 
@@ -433,6 +441,7 @@ int main(int argc, char *argv[])
     exit(EXIT_SUCCESS);
   }
 
+
   /*
    * If not in debugging mode, make myself a daemon 
    */
@@ -444,6 +453,16 @@ int main(int argc, char *argv[])
       fprintf(stderr, "Could not daemonize: %s\n", strerror(errno));
       exit(EXIT_FAILURE);
     }
+
+    if (pwent = getpwuid(getuid()))
+      username = pwent->pw_name;
+    /*
+     * The setup is finished, and we are daemonized, so we must now switch
+     * to syslog...
+     */
+    openlog("prngd", LOG_PID, LOG_DAEMON);
+    syslog(LOG_INFO, "prngd %s (%s) started up for user %s", PRNGD_VERSION,
+	   PRNGD_DATE, username);
   }
 
   /*
