@@ -1,9 +1,11 @@
+#include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <sys/time.h>
 #include <sys/times.h>
 #include <unistd.h>
 #include <fcntl.h>
+#include <syslog.h>
 
 #include "config.h"
 #include "prngd_version.h"
@@ -76,3 +78,8 @@ void SHA1_digest(SHA1_CTX *ctx, unsigned char *s);
 void SHA1_copy(SHA1_CTX *dest, SHA1_CTX *src);
 
 void close_on_exec(int fd, int on);
+
+void non_blocking(int fd, int on);
+
+ssize_t socket_read(int fd, void *buf, size_t num);
+ssize_t socket_write(int fd, void *buf, size_t num);
