@@ -7,10 +7,10 @@
 #include <signal.h>
 #include <time.h>
 #include <sys/stat.h>
-#include <sys/socket.h>
 #include <sys/un.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#include <sys/socket.h>
 #include <sys/wait.h>
 #include <netinet/in.h>
 
@@ -19,11 +19,13 @@
 static seed_t gatherer_seed;
 static pid_t gatherer_pid = -1;
 static int gatherer_status;
+static int caught_sigchild = 0;
+static volatile int received_sighup = 0;
 
 typedef enum
 {
   wait_command, wait_second_byte, wait_third_byte, wait_fourth_byte,
-  wait_read, wait_write, wait_entropy, to_be_closed
+  wait_read, wait_write, wait_entropy, to_be_closed, to_be_error_closed
 }
 querystate_t;
 
@@ -42,37 +44,34 @@ typedef struct
 static void termination_handler(int received_signal)
 {
   if (debug)
-    fprintf(stderr, "Caught SIG*, shutting down\n");
+    fprintf(stderr, "Caught SIGTERM, shutting down\n");
   close_seedfile();
   exit(EXIT_SUCCESS);
 }
 
+static void sighup_handler(int received_signal)
+{
+  received_sighup = 1;
+}
+
 static void child_handler(int received_signal)
 {
+  int save_errno = errno;
   pid_t pid;
 
-  if (debug)
-    fprintf(stderr, "Caught SIGCHLD\n");
+  /*
+   * Wait for child process to avoid zombies. This signal comes in
+   * asynchronously and only few functions are save to be called,
+   * especially stdio and syslog functions cannot be called.
+   */
+  caught_sigchild = 1;
 
   if ((pid = waitpid(-1, &gatherer_status, WNOHANG)) == -1)
-#ifdef PID_T_IS_LONG_INT
-    fprintf(stderr, "Couldn't wait for child %ld: %s\n", pid, strerror(errno));
-#else
-    fprintf(stderr, "Couldn't wait for child %d: %s\n", pid, strerror(errno));
-#endif
+    ;  /* Hmm, how to log this error without syslog... */
   else if ((pid == gatherer_pid) && WIFEXITED(gatherer_status))
      gatherer_pid = -1;
 
-  /*
-   * We should add the gatherer_status to the PRNG pool at this point, but
-   * we must not add entropy now because another rand_add() operation may be
-   * running just now. We hence save the data and add it later synchronized
-   * in the main_loop(). We don't need to care about the contents of
-   * gatherer_status et al and what happens if the connection is shut down
-   * from PRNGD because too many bytes were sent, since we don't care about
-   * the data at all, we just want to add to the pool as much as we can.
-   */
-  seed_internal(&gatherer_seed);
+  errno = save_errno;
 }
 
 void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
@@ -109,8 +108,9 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
   gather_start = time(NULL);	/* Initialize to make gcc happy */
 
   /*
-   * Set up the signal handling. TERM and HUP signals shall lead to clean
+   * Set up the signal handling. TERM signals shall lead to clean
    * shutdown of prngd (save the seed and exit).
+   * HUP signals shall lead to a re-read of the config file.
    * SIGPIPE might occur when a peer shuts down the connection. Of course,
    * we don't want prngd to be affected in any way (it should simply close
    * the connection), so SIGPIPE shall be ignored.
@@ -120,6 +120,11 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
   sigemptyset(&new_action.sa_mask);
   new_action.sa_flags |= SA_RESTART;
   sigaction(SIGTERM, &new_action, NULL);
+
+  memset(&new_action, 0, sizeof(struct sigaction));
+  new_action.sa_handler = sighup_handler;
+  sigemptyset(&new_action.sa_mask);
+  new_action.sa_flags |= SA_RESTART;
   sigaction(SIGHUP, &new_action, NULL);
 
   memset(&new_action, 0, sizeof(struct sigaction));
@@ -130,6 +135,8 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 
   memset(&new_action, 0, sizeof(struct sigaction));
   new_action.sa_handler = SIG_IGN;
+  sigemptyset(&new_action.sa_mask);
+  new_action.sa_flags |= SA_RESTART;
   sigaction(SIGPIPE, &new_action, NULL);
 
   for (;;)
@@ -181,6 +188,18 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
     }
     num = select(max_fd + 1, &rfds, &wfds, &efds, &timeout);
 
+    if (caught_sigchild) {
+      caught_sigchild = 0;
+      if (debug)
+	fprintf(stderr, "Caught SIGCHLD\n");
+    }
+
+    if (received_sighup) {
+      received_sighup = 0;
+      if (debug)
+	fprintf(stderr, "Received SIGHUP\n");
+    }
+
     if (time(0) >= last_seed_stat + seed_stat_interval) {
       seed_stat();
       last_seed_stat = time(0);
@@ -198,7 +217,7 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 	 * Bytes are coming back from the gatherer process. Read chunks of
 	 * up to 10240 bytes and feed them into the PRNG
 	 */
-	numbytes = read(gather_fd, gatherbuf, 10240);
+	numbytes = socket_read(gather_fd, gatherbuf, 10240);
 	rand_add(gatherbuf, numbytes, est * numbytes);
 	if (debug && numbytes > 0)
 	  fprintf(stderr, "Read %d bytes from gatherer\n", numbytes);
@@ -305,18 +324,7 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 	 */
 	close(pipedes[1]);
 	gather_fd = pipedes[0];
-        if ((flags = fcntl(gather_fd, F_GETFL, 0)) < 0) {
-	  if (debug)
-	    fprintf(stderr, "Could not get socket file mode: %s\n",
-		  strerror(errno));
-	  close(gather_fd);
-        }
-        if (fcntl(gather_fd, F_SETFL, flags | O_NONBLOCK) < 0) {
-	  if (debug)
-	    fprintf(stderr, "Could not set socket file mode: %s\n",
-		  strerror(errno));
-	  close(gather_fd);
-        }
+	non_blocking(gather_fd, 1);
 	gather_start = time(NULL);
 	fprintf(stderr, "Spawned gatherer: %s (%s",
 		entropy_source[gatherer_entry].path,
@@ -332,7 +340,7 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
     /*
      * Do service the active queries first, because accepting a new query
      * will create a new fd. First step: check errors reported and mark
-     * the corresponding descriptor "to_be_closed".
+     * the corresponding descriptor "to_be_error_closed".
      */
     for (i = 0; i < max_query; i++)
     {
@@ -340,7 +348,7 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
       {
 	if (debug)
 	  fprintf(stderr, "Error flagged for service fd %d\n", egd_query[i].fd);
-	egd_query[i].state = to_be_closed;
+	egd_query[i].state = to_be_error_closed;
       }
       else if (FD_ISSET(egd_query[i].fd, &rfds) ||
 	       FD_ISSET(egd_query[i].fd, &wfds) ||
@@ -354,13 +362,13 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 	    egd_query[i].state = wait_write;
 	break;
       case wait_command:
-	error = read(egd_query[i].fd, &(egd_query[i].command), 1);
-	if ((error == -1) && (errno == EAGAIN))
-	  break;
-	if (error != 1)
+	error = socket_read(egd_query[i].fd, &(egd_query[i].command), 1);
+	if (error == -1)
+	  egd_query[i].state = to_be_error_closed;
+	else if (error != 1)
 	  egd_query[i].state = to_be_closed;
 	else if (egd_query[i].command > 4)
-	  egd_query[i].state = to_be_closed;
+	  egd_query[i].state = to_be_error_closed;
 	else if (egd_query[i].command == 0)
 	{
 	  numbytes = rand_pool();
@@ -390,9 +398,9 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 	}
 	break;
       case wait_second_byte:
-	error = read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
+	error = socket_read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
 	if (error != 1)
-	  egd_query[i].state = to_be_closed;
+	  egd_query[i].state = to_be_error_closed;
 	else if (egd_query[i].command == 1)
 	{
 	  seed_internal(NULL);	/* Mix in just before query */
@@ -424,9 +432,9 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 	}
 	break;
       case wait_third_byte:
-	error = read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
+	error = socket_read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
 	if (error != 1)
-	  egd_query[i].state = to_be_closed;
+	  egd_query[i].state = to_be_error_closed;
 	else
 	{
 	  egd_query[i].entropy_bits += egd_query[i].secbyte;
@@ -434,9 +442,9 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 	}
 	break;
       case wait_fourth_byte:
-	error = read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
+	error = socket_read(egd_query[i].fd, &(egd_query[i].secbyte), 1);
 	if (error != 1)
-	  egd_query[i].state = to_be_closed;
+	  egd_query[i].state = to_be_error_closed;
 	else
 	{
 	  egd_query[i].to_process = egd_query[i].secbyte;
@@ -446,9 +454,9 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 	break;
       case wait_read:
 	num = egd_query[i].to_process - egd_query[i].yet_processed;
-	error = read(egd_query[i].fd, egd_query[i].buffer, num);
+	error = socket_read(egd_query[i].fd, egd_query[i].buffer, num);
 	if (error <= 0)
-	  egd_query[i].state = to_be_closed;
+	  egd_query[i].state = to_be_error_closed;
 	else
 	{
 	  rand_add(egd_query[i].buffer, error,
@@ -460,12 +468,12 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 	}
 	break;
       case wait_write:
-	error = write(egd_query[i].fd,
+	error = socket_write(egd_query[i].fd,
 		      &(egd_query[i].buffer[egd_query[i].yet_processed]),
 		      egd_query[i].to_process -
 		      egd_query[i].yet_processed);
-	if (error < 1)
-	  egd_query[i].state = to_be_closed;
+	if (error <= 0)
+	  egd_query[i].state = to_be_error_closed;
 	else
 	  egd_query[i].yet_processed += error;
 	if (egd_query[i].yet_processed == egd_query[i].to_process) {
@@ -480,9 +488,14 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
       }
     }
     for (i = 0; i < max_query; i++) {
-      if (egd_query[i].state == to_be_closed) {
+      if ((egd_query[i].state == to_be_closed) ||
+	  (egd_query[i].state == to_be_error_closed)) {
 	if (debug)
-	  fprintf(stderr, "Closing service fd %d\n", egd_query[i].fd);
+	  fprintf(stderr, "Closing service fd %d, error=%d\n",
+			  egd_query[i].fd,
+			  (egd_query[i].state == to_be_error_closed));
+	else if (egd_query[i].state == to_be_error_closed)
+	  syslog(LOG_INFO, "closing service fd %d for error.", egd_query[i].fd);
 	close(egd_query[i].fd);
 	memset(egd_query[i].buffer, 0, 256);	/* Remove delivered entropy */
 	if (i < max_query - 1)
@@ -507,22 +520,7 @@ void main_loop(int *service_socket, int numsock, struct sockaddr_un sockun,
 				 * Ignore and stay alive 
 				 */
 
-        if ((flags = fcntl(accept_fd, F_GETFL, 0)) < 0)
-        {
-	  if (debug)
-	    fprintf(stderr, "Could not get socket file mode: %s\n",
-		    strerror(errno));
-	  close(accept_fd);
-	  continue;
-        }
-        if (fcntl(accept_fd, F_SETFL, flags | O_NONBLOCK) < 0)
-        {
-	  if (debug)
-	    fprintf(stderr, "Could not set socket file mode: %s\n",
-		    strerror(errno));
-	  close(accept_fd);
-	  continue;
-        }
+	non_blocking(accept_fd, 1);
 	close_on_exec(accept_fd, 1);
 
         /*
